#ifndef ___SC2Map_hpp___
#define ___SC2Map_hpp___

#include <vector>
#include <list>
#include <map>
#include <set>
#include <string>
using namespace std;

#include "StormLib.h"
#include "tinyxml.h"
#include "pngwriter.h"

#include "sc2mapTypes.hpp"
#include "coordinates.hpp"
#include "PrioQueue.hpp"


class SC2Map
{

public:

  //////////////////////////////////////////////////
  // implemented in SC2Map.cpp
  //////////////////////////////////////////////////
  SC2Map();
  ~SC2Map();


  //////////////////////////////////////////////////
  // implemented in read.cpp
  //////////////////////////////////////////////////
  int readMap( const char* strArchiveFilename );


  //////////////////////////////////////////////////
  // implemented in bases.cpp
  //////////////////////////////////////////////////
  static void setInfluenceWeightGround( float iw );
  static void setInfluenceWeightCWalk ( float iw );
  static void setInfluenceWeightAir   ( float iw );

  static void setOpennessNeighborhoodRadius( float onr );

  void identifyBases();
  void analyzeBases();


  //////////////////////////////////////////////////
  // implemented in openness.cpp
  //////////////////////////////////////////////////
  static void setOpennessRenderMax( float orm );

  void computeOpenness();


  //////////////////////////////////////////////////
  // implemented in dijkstra.cpp
  //////////////////////////////////////////////////
  void prepShortestPaths();


  //////////////////////////////////////////////////
  // implemented in spreadsheet.cpp
  //////////////////////////////////////////////////
  void writeToCSV( const char* strOutCSV );


  //////////////////////////////////////////////////
  // implemented in render.cpp
  //////////////////////////////////////////////////

  // one image, uses path type="ground with rocks"
  void renderImageOpenness();

  // two images for each start location pair,
  // one for paths between mains, one between nats
  void renderImagesShortestPaths();

  // one image per pair of start locations
  void renderImagesInfluence();

  // one image, just terrain layer
  void renderImageTerrainOnly();

  // one image for each pathing type
  void renderImagesPathing();

  // one image, all bases with labels
  void renderImageBases();

  // for development
  void renderTestImages();


// TODO: UNDO THIS LATER
// really don't want to write all the accessors
// for the SC2MapAggregator right now, so make
// all the members of this class public
//protected:


  //////////////////////////////////////////////////
  // in read.cpp
  //////////////////////////////////////////////////
  int readArchiveFile( const HANDLE archive,
                       const char*  strFilename,
                       int*         bufferOutSize,
                       u8**         bufferOut );

  int verifyMagicWord( const u8* magic, const u8* buffer );

  int readGameStrings_txt( const HANDLE archive );
  int readMapInfo        ( const HANDLE archive );
  int readt3HeightMap    ( const HANDLE archive );
  int readt3CellFlags    ( const HANDLE archive );
  int readObjects        ( const HANDLE archive );

  // put the raw string of the map in strMapName and put
  // a cleaned up version of that string in strMapFilename
  // for opening output files with valid filenames
  char strArchiveName[FILENAME_LENGTH];
  char strMapName    [FILENAME_LENGTH];
  char strMapFilename[FILENAME_LENGTH];

  void setMapNameFromArchiveName( const char* strArchiveFilename );
  void ensureMapNameValidFilename();

  // when opening several maps, warn when a map filename is reused
  static set<string> mapFilenamesUsed;


  // note that the dimensions of the total map
  // and playable area are not given in points
  // because it is counter-intuitive to access
  // something like tSizeMap.mtx.  Other
  // coordinates make sense as points.

  // total map dimensions in terrain units
  // and cell units
  int txDimMap;
  int tyDimMap;
  int cxDimMap;
  int cyDimMap;

  // playable area dimensions
  int txDimPlayable;
  int tyDimPlayable;
  int cxDimPlayable;
  int cyDimPlayable;


  // boundaries of playable area, note that
  // left, bottom is inclusive but right, top
  // is not
  point cLeftBottom;
  point cRightTop;

  point tLeftBottom;
  point tRightTop;


  inline bool isPlayableTerrain( point* t )
  {
    return
      t->ptx >= 0 && t->ptx < txDimPlayable &&
      t->pty >= 0 && t->pty < tyDimPlayable;
  }


  inline bool isPlayableCell   ( point* c )
  {
    return
      c->pcx >= 0 && c->pcx < cxDimPlayable &&
      c->pcy >= 0 && c->pcy < cyDimPlayable;
  }


  // the available cliff heights: 0 is unplayable
  // low ground and levels 1, 2, 3 may be in play
  u8*  mapHeight;
  u8   getHeight( point* t );
  void setHeight( point* t, u8 h );


  // the pathing map is 0 (false) if no pathing is
  // allowed in the cell and 1 (true) if allowed--
  // build pathing maps for cliff walkers also, and
  // do a version of paths with and without
  // destructible rocks
  bool* mapPathing;
  bool  getPathing( point* c, PathType t );
  void  setPathing( point* c, PathType t, bool p );

  // calculate this when reading in the map data
  // and make it available for analysis
  int numPathableCells[NUM_PATH_TYPES];



  //////////////////////////////////////////////////
  // in placedobjects.cpp
  //////////////////////////////////////////////////
  list<StartLoc*> startLocs;

  static float defaultMineralAmount;
  static float defaultGeyserAmount;

  float totalMinerals;
  float totalVespeneGas;
  float totalHYMinerals;
  float totalHYVespeneGas;

  list<Resource*> resources;

  void processPlacedObject( TiXmlElement* object );

  void processDestructible1x1( point* p );
  void processDestructible2x2( point* p );
  void processDestructible4x4( point* p );
  void processDestructible6x6( point* p );
  void processDestructible2x4Horizontal( point* p );
  void processDestructible2x4Vertical  ( point* p );
  void processDestructible2x6Horizontal( point* p );
  void processDestructible2x6Vertical  ( point* p );
  void processDestructibleHugeBLUR     ( point* p );
  void processDestructibleHugeULBR     ( point* p );

  void removePathing( point* c );
  void removePathing( point* c, PathType t );



  //////////////////////////////////////////////////
  // in bases.cpp
  //////////////////////////////////////////////////
  static float influenceWeightGround;
  static float influenceWeightCWalk;
  static float influenceWeightAir;

  static float opennessNeighborhoodRadius;

  list<Base*> bases;
  set <Base*> islandExpos;
  set <Base*> semiislandExpos;

  void locateBases();
  void nameStartLocs();
  void calculateAverageOpenness();
  void calculateInfluence();
  void assignMainNatThirdIslands();
  void calculatePositionalBalance();

  bool getNearestBase( Resource* r, Base** bOut );

  float weightedInfluenceDistance( StartLoc* sl, Base* b );
  float averageInfluence         ( StartLoc* sl, Base* b );



  //////////////////////////////////////////////////
  // in openness.cpp
  //////////////////////////////////////////////////
  static float OPENNESS_NOTCALCULATED;
  static float opennessRenderMax;

  // openness is a numerical value for a cell that
  // determines how far away the nearest unpathable
  // cell is, or how "open" the cell is
  float* mapOpenness;
  float* mapOpennessPrev;

  void computeOpenness( PathType t );

  bool  checkHasOpenness( point* c, PathType t );
  float getOpenness     ( point* c, PathType t );
  void  setOpenness     ( point* c, PathType t, float o );

  bool  checkHasOpennessLastPass( point* c, PathType t );
  float getOpennessLastPass     ( point* c, PathType t );

  // max openness values found for the map, one for
  // each pathing type
  float opennessMax[NUM_PATH_TYPES];



  //////////////////////////////////////////////////
  // in dijkstra.cpp
  //////////////////////////////////////////////////

  // for connecting the pathing graphs
  static float k0; // distance u-4, etc.
  static float k1; // distance u-8, etc.
  static float neighborWeights[NUM_NODE_NEIGHBORS];

  void buildPathGraph( PathType t );
  void computeShortestPaths( Node* src, PathType t );

  // there is a disjoint graph of nodes for each path
  // type all contained in this array
  Node** mapPathNodes;
  Node*  getPathNode( point* c, PathType t );
  void   setPathNode( point* c, PathType t, Node* u );

  // for each path type there is a sequence of nodes
  // where each node is also in the map of path nodes
  // by location, and the node's position in this vector
  // is the same as its ID
  vector<Node*> nodes[NUM_PATH_TYPES];

  // the shortest path distance (d) and
  // predecessor (pi) results of many-pairs shortest
  // path calculations.  When you need to access say:
  //   *shortest path distance, d
  //   *for path type t
  //   *from node ID=i to node ID=j you would use:
  //
  //     (*(d[t][i]))[j]
  //
  map< int, vector<float>* > d [NUM_PATH_TYPES];
  map< int, vector<Node*>* > pi[NUM_PATH_TYPES];

  // best used by other modules--if you ask for shortest distance
  // from points that are out of bounds or over unpathable cells
  // you get a nice meaningful infinity returned
  float getShortestPathDistance( point* src, point* dst, PathType t );

  float getShortestPathDistance   ( Node* u, Node* v, PathType t );
  void  setShortestPathDistance   ( Node* u, Node* v, PathType t, float dIn );

  Node* getShortestPathPredecessor( Node* u, Node* v, PathType t );
  void  setShortestPathPredecessor( Node* u, Node* v, PathType t, Node* piIn );

  // a reusable priority queue for the shortest path
  // algorithm (Dijkstra's)
  PrioQueue pqueue;

  // doesn't require Dijkstra's, but fits nicely in this file anyway
  float getShortestAirDistance( point* src, point* dst );


  // just a helper func
  static bool effectivelyInfinity( float x );



  //////////////////////////////////////////////////
  // implemented in spreadsheet.cpp
  //////////////////////////////////////////////////




  //////////////////////////////////////////////////
  // in render.cpp
  //////////////////////////////////////////////////
  static int iLeft;
  static int iBottom;
  static int iRight;
  static int iTop;
  static int iDimT;
  static int iDimC;

  int ix2png( int ix );
  int iy2png( int iy );

  void resetPNGWriter( const char* strFileSuffix,
                       int         iBorderLeft,
                       int         iBorderRight,
                       int         iBorderTop,
                       int         iBorderBottom );

  void renderFooter();

  void renderTerrain();
  void plotTerrain( point* t );

  void renderImagePathing( const char* strOutPNG, PathType t );
  void renderPathing( PathType t );
  void plotPathing( point* c, PathType t );

  void renderOpenness( PathType t );
  void plotOpenness( point* c, PathType t );

  void openness2color( float oIn,
                       float* rOut, float* gOut, float* bOut );

  void renderResources();
  void plotMineralField( point* p, bool highYield );
  void plotGeyser( point* p );

  void renderStartLocs();
  void plotStartLoc( StartLoc* sl );

  void renderBases();
  void plotBase( Base* b );

  void renderImageInfluence( StartLoc* sl1, StartLoc* sl2 );
  void renderInfluence     ( StartLoc* sl1, StartLoc* sl2 );

  void renderImageShortestPaths( char*  filename,
                                 char*  title,
                                 point* p0,
                                 point* p1 );

  void plotAirPath( point* p0, point* p1,
                    float r, float g, float b,
                    float labelPlacement );

  void plotShortestPath( point* p0, point* p1, PathType t,
                         float r, float g, float b,
                         float labelPlacement );

  void plotAllShortestPaths( point* src, PathType t );

  void plotTerrainPoint( point* t, float r, float g, float b );
  void plotCellPoint   ( point* c, float r, float g, float b );

  void plotArrow( int x0, int y0, int x1, int y1,
                  float r, float g, float b );

  void plotLine( int x0, int y0, int x1, int y1,
                 float thick,
                 float r, float g, float b );

  void plotTri( int x1, int y1,
                int x2, int y2,
                int x3, int y3,
                float r, float g, float b );

  void plotText( char* text,
                 int fontsize,
                 bool centered, int ix, int iy,
                 float angle,
                 float rf, float gf, float bf, // foreground
                 float rb, float gb, float bb  // background
               );
  int getTextWidth( char* text, int fontsize );

  void renderTestFrameTranslation();
  void renderTestClockFaceDetection();
  void renderTestAllPaths();

  pngwriter* img;
};


#endif // ___SC2Map_hpp___
